# Note: implement power series rings with finite precision as
# quotients of polynomial rings.

if 'ATTACH' not in globals():
    from finite_algebra import bilinear_map_to_algebra

class BundledCurve(SageObject):
    """
    Construct a curve over a field with a given line bundle on it.

    INPUT:
     - ``ring`` - a finite algebra over a field
     - ``basis_V1`` - a list of elements of ``ring``

    Let ``C`` be a (projective, smooth, geometrically connected) curve
    of genus ``g`` over a field ``k``.  Let ``L`` be a line bundle on
    ``C`` such that the degree of ``L`` is at least 2g + 1.  We assume
    that an effective divisor ``D`` on ``C`` has been chosen.  Let
    ``R`` be the coordinate ring of ``D``.  We assume a trivialisation
    of the restriction of ``L`` to ``D`` has been chosen.  For every
    integer ``i``, the induced trivialisation of the ``i``-th tensor
    power ``L^i`` identifies the space of global sections of ``L^i``
    with a ``k``-linear subspace ``V_i`` of ``R``.
    """
    def __init__(self, ring, basis_V1):
        self._base_field = ring.base_ring()
        self._ring = ring
        self._V = [[self._ring(1)], basis_V1]
        self._V_ring_vectors = [Matrix ([self._section_to_ring_vector(s)
                                         for s in V])
                                for V in self._V]
        self._igs = {}

    def base_ring(self):
        return self._base_field

    def _section_to_ring_vector(self, s):
        return vector(s)

    def _ring_vector_to_section(self, v):
        return self._ring(list(v))

    def _section_to_vector(self, s, i):
        self._extend_sections(i)
        Vi = self._V_ring_vectors[i]
        return Vi.solve_left(self._section_to_ring_vector(s))

    def _sections_to_matrix(self, S, i):
        return Matrix([self._section_to_vector(s, i) for s in S])

    def _vector_to_section(self, v, i):
        self._extend_sections(i)
        Vi = self._V_ring_vectors[i]
        return self._ring_vector_to_section(v * Vi)

    def _matrix_to_sections(self, M, i):
        return [self._vector_to_section(v, i) for v in M.rows()]

    def _multiply_subspaces(self, W1, W2):
        P = [self._section_to_ring_vector(s * t) for s in W1 for t in W2]
        M = Matrix(P)
        return [self._ring_vector_to_section(v) for v in M.row_space().basis()]

    def _extend_sections(self, i):
        for j in range(len(self._V), i + 1):
            self._V.append(self._multiply_subspaces(self._V[1],
                                                    self._V[j - 1]))
            self._V_ring_vectors.append(
                  Matrix([self._section_to_ring_vector(s)
                          for s in self._V[j]]))

    def sections(self, i):
        """
        Return the space of global sections of ``L^i``.

        INPUT:
         - ``i`` - an integer

        OUTPUT:
         - list of RingElements - a basis for the space of global
           sections of ``L^i``, represented as elements of ``R`` via
           the trivialisation of the restriction of ``L`` to ``D``.

        EXAMPLES: TODO
        """
        self._extend_sections(i)
        return self._V[i]

    def random_section(self, i):
        """
        Return a random section of H^0(L^i)

        INPUT:
         - ``i`` - an integer

        OUTPUT: RingElement - see self.sections(i)
        """
        result = self._ring(0)
        k = self.base_ring()
        for section in self.sections(i):
            result += k.random_element() * section
        return result

    def zero_divisor(self, i=1):
        return BundledCurveEffectiveDivisor(self, i, self.sections(i))

    def random_divisor(self, d, tries=Infinity):
        """
        Return a random effective divisor of degree ``d`` on self.

        INPUT:
         - ``d`` - integer
         - ``tries`` - integer or Infinity (default: Infinity)

        OUTPUT: BundledCurveEffectiveDivisor
        """
        g = self.genus()
        dl = self.degree()
        i = ((d + 2*g + 1) / dl).ceil()
        while tries > 0:
            tries -= 1
            s = self.random_section(i)
            mat_s = self.multiplication_matrix(i, s, 1)
            D = BundledCurveEffectiveDivisor(self, i+1, mat_s)
            primes = D.decompose()
            primes = [(prime[0], prime[0].degree()) for prime in primes]
            primes = [prime for prime in primes if prime[1] <= d]
            # TODO: something smarter than all subsets:
            subset_degrees = [(subset, sum([primes[k][1] for k in subset])) for
                              subset in subsets(range(len(primes)))]
            deg_d_subsets = [subset[0] for subset in subset_degrees if
                             subset[1] == d]
            if deg_d_subsets == []:
                continue
            deg_d_subset = deg_d_subsets[randint(0, len(deg_d_subsets) - 1)]
            result = self.zero_divisor()
            for k in deg_d_subset:
                result = (result + primes[k][0]).simplify()
            return result
        raise ArithmeticError, "cannot find divisor of degree %s" % d

    def sections_igs(self, i):
        """
        Return an ideal generating set for H^0(L^i).

        INPUT:
         - ``i`` - an integer

        OUTPUT:
         - list of ring elements, representation as in self.sections(i).
           The common divisor of these elements is 0.  This is mainly used
           to speed up computations as an igs is usually very small
        """
        if i not in self._igs:
            result = []
            d = len(self.sections(2 * i + 1))
            V = self.base_ring() ^ d
            W = V.zero_subspace()
            while W.dimension() < d:
                s = self.random_section(i)
                result.append(s)
                W += self.multiplication_matrix(i, s, i + 1).row_space()
            self._igs[i] = result
        return self._igs[i]

    def multiplication_matrix(self, i, s, j):
        """
        Return the matrix of multiplication by an element ``s`` of
        ``H^0(L^i)`` as a ``k``-linear map from ``H^0(L^j)`` to
        ``H^0(L^{i+j})``, with respect to fixed bases of these spaces.

        INPUT:
         - ``i`` - integer
         - ``s`` - RingElement
         - ``j`` - integer

        OUTPUT: matrix
        """
        Vj = self.sections(j)
        Vij = self.sections(i + j)
        result = []
        for t in Vj:
            result.append(self._section_to_vector(s * t, i + j))
        return matrix(result)

    def degree(self):
        return len(self.sections(2)) - len(self.sections(1))

    def genus(self):
        return len(self.sections(2)) - 2 * len(self.sections(1)) + 1

    def divide_subspaces(self, V, i, S, j):
        """
        Divide the subspace ``V`` of H^0(L^i) by the sections S of H^0(L^j).

        INPUT:
         - V - matrix representing vector space
         - i - integer
         - S - list of RingElements - see self.sections(j)
         - j - integer

        OUTPUT: VectorSpace

        The output vector space is the space {t in H^0(L^{i-j}):
        st in V for all s in S}.
        """
        V_ker_mat = V.echelon_form().right_kernel().matrix().transpose()
        big_mat = [self.multiplication_matrix(j, s, i - j) * V_ker_mat for s in S]
        mat = reduce(lambda x, y: x.augment(y), big_mat)
        return mat.kernel()

    def distinguished_point(self):
        """
        Return a rational point on the curve.

        OUTPUT: BundledCurveEffectiveDivisor

        This method raises a NotImplementedError, but may be implemented
        in subclasses.  A distinguished point can be used, for instance,
        to embed self into its Jacobian.
        """
        raise NotImplementedError

    def distinguished_point_multiple(self, i):
        """
        Return a specified multiple of the distinguished point.
        
        INPUT:
         - i - non-negative integer
        
        OUTPUT: BundledCurveEffectiveDivisor
        
        This method returns ``i`` times the distinguished point,
        represented as an effective divisor.  It depends on the method
        distinguished_point being implemented.
        """
        try:
            return self._mult_dist_point[i]
        except AttributeError:
            self._mult_dist_point = [self.zero_divisor()]
        except IndexError:
            pass
        for j in range(len(self._mult_dist_point) - 1, i):
            self._mult_dist_point.append((self.distinguished_point() + self._mult_dist_point[j]).simplify())
        return self._mult_dist_point[i]
    
    def __call__(self, x):
        """
        If self is C.base_extend(bla) and x is on C, extend it to self.

        INPUT:
         - ``x`` - BundledCurveEffectiveDivisor or
                   BundledCurveRationalFunction

        OUTPUT: BundledCurveEffectiveDivisor or
                BundledCurveRationalFunction (same type as ``x``)
        """
        if isinstance(x, BundledCurveEffectiveDivisor):
            return BundledCurveEffectiveDivisor(self, divisor.i(),
                        x.basis_matrix().base_extend(self.base_ring()))
        if isinstance(x, BundledCurveRationalFunction):
            return BundledCurveRationalFunction(self, x.i(),
                        x.sections()[0].base_extend(self._ring),
                        x.sections()[1].base_extend(self._ring))
        raise TypeError


class BundledCurveEffectiveDivisor(SageObject):
    """
    Construct an effective divisor on the given curve.

    INPUT:
     - ``curve`` - BundledCurve
     - ``i`` - integer
     - ``W`` - list of RingElements, or matrix

    We represent an effective divisor ``D`` by the space ``W`` =
    ``H^0(L^i(-D))``, for ``i`` sufficiently large such that ``i deg L
    - deg D`` is at least 2g + 1.
    """
    def __init__(self, curve, i, W):
        self._curve = curve
        self._i = i
        if isinstance(W, list):
            self._W = W
            self._W_matrix = curve._sections_to_matrix(W, i)
        else:
            self._W_matrix = W
            self._W = curve._matrix_to_sections(W, i)

    def _repr_(self):
        return "Effective divisor on %s" % self.curve()

    def curve(self):
        return self._curve

    def base_ring(self):
        return self.curve().base_ring()

    def i(self):
        return self._i

    def basis(self):
        return self._W

    def basis_matrix(self):
        return self._W_matrix

    def igs(self):
        raise NotImplementedError

    def igs_matrix(self):
        raise NotImplementedError

    def degree(self):
        return len(self.curve().sections(self.i())) - len(self.basis())

    def __add__(self, other):
        # TODO: this is too slow, make it faster
        C = self.curve()
        if C is not other.curve():
            raise ValueError, "divisors not on the same curve"
        i = self._i + other._i
        W = C._multiply_subspaces(self.basis(), other.basis())
        return BundledCurveEffectiveDivisor(C, i, W)

    def simplify(self, i=None):
        """
        Output the divisor ``D`` represented by self, as H^0(L^i(-D)).

        INPUT:
         - ``i`` - integer or None (default: None)

        OUTPUT: BundledCurveEffectiveDivisor

        If i is None, we take the smallest i with deg L^i(-D) >= 2g.

        EXAMPLES: TODO
        """
        C = self.curve()
        if i is None:
            d = self.degree()
            dl = C.degree()
            g = C.genus()
            i = ((d + 2*g) / dl).ceil()
        ii = self.i()
        if i == ii:
            return self
        elif i < ii:
            V = C.divide_subspaces(self.basis_matrix(), ii,
                                   C.sections_igs(ii - i), ii - i)
            return BundledCurveEffectiveDivisor(self.curve(), i, V.matrix())
        else:
            return self + BundledCurveEffectiveDivisor(C, i - ii,
                                                       C.sections(i - ii))

    def algebra(self):
        """
        Compute the coordinate ring of self, seen as subscheme of self.curve().

        OUTPUT: a tuple (A, m) consisting of:
         - ``A`` - FiniteAlgebra
         - ``m`` - matrix - determines quotient map H^0(L^i) -> A

        EXAMPLES: TODO
        """
        i = self.i()
        C = self.curve()
        W = self.basis_matrix().echelon_form().row_space()
        W_compl, m, W_lift = W.ambient_vector_space().quotient_abstract(W)
        W2 = self.simplify(2 * i).basis_matrix().echelon_form().row_space()
        W2_compl, m2, W2_lift = W2.ambient_vector_space().quotient_abstract(W2)
        bil_map = []
        for w in W_compl.basis():
            w1 = W_lift(w)
            s1 = C._vector_to_section(w1, i)
            w1_matrix = C.multiplication_matrix(i, s1, i)
            bil_map.append(W_lift.matrix() * w1_matrix * m2.matrix())
        A = bilinear_map_to_algebra(self.base_ring(), bil_map)
        return (A, m.matrix())

    def decompose(self):
        """
        Decompose self into prime divisors.

        OUTPUT: list of tuples (P, e), with:
         - ``P`` - BundledCurveEffectiveDivisor
         - ``e`` - integer - indicates multiplicity

        EXAMPLES: TODO
        """
        A, m = self.algebra()
        result = []
        for phi in A.primary_decomposition():
            B = phi.target()
            max_id_B = B.maximal_ideal()
            max_id = phi.inverse_image(max_id_B)
            # Compute inverse image of max_id along m:
            coker_max_id = max_id.basis_matrix().transpose().kernel().basis_matrix().transpose()
            inverse_image = (m * coker_max_id).kernel().basis_matrix()
            P = BundledCurveEffectiveDivisor(self.curve(), self.i(), inverse_image)
            # Determine multiplicity:
            dim_B = B.degree()
            dim_max_id = max_id_B.basis_matrix().rank()
            e = dim_B // (dim_B - dim_max_id)
            result.append((P, e))
        return result

    def frobenius(self, n=1):
        """
        Apply the absolute Frobenius Frob_p^n to self.

        INPUT:
         - ``n`` - integer (default: 1)

        OUTPUT: BundledCurveEffectiveDivisor

        EXAMPLES: TODO
        """
        if self.base_ring().is_prime_field():
            return self
        matrix = self.basis_matrix().apply_map(lambda x: x.frobenius(n))
        return BundledCurveEffectiveDivisor(self.curve(), self.i(), matrix)

    def normalise(self):
        """
        Find the O-normalised effective divisor attached to the
        line bundle L^i(-self), where i = self.i() and O is the
        distinguished rational point.  This is an effective divisor
        of degree equal to the genus of the curve.

        OUTPUT: BundledCurveEffectiveDivisor

        EXAMPLES: TODO
        """
        X = self.curve()
        i = self.i()
        # Find the largest j for which L^i(-self - jO) has a non-zero
        # global section s.  Such an s is unique up to scaling; in
        # particular, its divisor is unique.
        for j in range(i * X.degree() - self.degree(), -1, -1):
            D = self + X.distinguished_point_multiple(j)
            # Note that the usual degree bounds may not be satisfied
            # in the following line.  This means that D.simplify(i)
            # may define a divisor of the form D + E where E is
            # an effective divisor; this is precisely what we are
            # interested in.
            s = D.simplify(i).basis()
            if len(s) == 1:
                break
        ii = D.i()
        sV = X._multiply_subspaces(s, X.sections(ii))
        W = X.divide_subspaces(X._sections_to_matrix(sV, i + ii), i + ii,
                               D.basis(), ii)
        E = BundledCurveEffectiveDivisor(X, i, W.basis_matrix())
        return (E + X.distinguished_point_multiple(X.genus() - E.degree())).simplify()


class BundledCurveRationalFunction(SageObject):
    """
    Construct a rational function on a BundledCurve.
    
    INPUT:
     - ``C`` - BundledCurve
     - ``i`` - integer
     - ``s`` - RingElement - section of H^0(C, L^i)
     - ``t`` - RingElement - section of H^0(C, L^i)
    
    This defines the function s/t on C.
    
    EXAMPLES: TODO
    """
    
    def __init__(self, C, i, s, t):
        self._curve = C
        self._i = i
        self._sections = (s, t)
        
    def _repr_(self):
        return "Rational function on %s" % self.curve()
    
    def curve(self):
        return self._curve
    
    def sections(self):
        return self._sections
    
    def i(self):
        return self._i
    
    def base_ring(self):
        return self.curve().base_ring()
    
    def __call__(self, D):
        """
        Evaluate self at an effective divisor.
        
        INPUT:
         - ``D`` - BundledCurveEffectiveDivisor - one on self.curve()
        
        OUTPUT: polynomial over the base field.
        
        If D is a sum of points P_i, return the monic polynomial whose
        roots are the s(P_i)/t(P_i), counted with multiplicities.
        
        EXAMPLES: TODO
        """
        C = self.curve()
        if C is not D.curve():
            raise ValueError, "divisor not on the same curve as function"
        s, t = self.sections()
        D = D.simplify()
        if D.i() < self.i():
            D = D.simplify(self.i())
        elif D.i() > self.i():
            u = C.sections(D.i() - self.i())[0]
            s *= u
            t *= u
            # TODO: it might happen that u is not invertible at D, fix this
        A, m = D.algebra()
        st = C._sections_to_matrix([s, t], D.i())
        s, t = [A(x).matrix() for x in list(st * m)]
        try:
            return (s * ~t).charpoly()
        except ZeroDivisionError:
            raise ZeroDivisionError, "denominator is not invertible at D"
