class FiniteAlgebra(Ring):
    """
    Create a finite commutative k-algebra from a multiplication table.

    INPUT:
     - ``k`` - a field
     - ``table`` - a list of matrices

    The list ``table`` must have the following form: there exists a
    finite commutative ``k``-algebra of degree ``n`` with basis
    ``(b_1,...,b_n)`` such that the ``i``-th element of ``table`` is
    the matrix of multiplication by ``b_i`` with respect to the basis
    ``(b_1,...,b_n)``.

    EXAMPLES::

        sage: A = FiniteAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]), Matrix([[0, 1], [0, 0]])])
        sage: A
        Finite algebra of degree 2 over Finite Field of size 3
        sage: A.degree()
        2
        sage: A.base_ring()
        Finite Field of size 3
    """

    def __init__(self, k, table):
        Ring.__init__(self, base=k)
        self._table = [b.base_extend(k) for b in table]
        n = len(table)
        for b in table:
            if b.nrows() != n or b.ncols() != n:
                raise ValueError, "Input is not a multiplication table"
        self._degree = n
        B = reduce(lambda x, y: x.augment(y), self._table, Matrix(k, n, 0))
        # This is the vector obtained by concatenating the rows of the
        # n times n identity matrix.
        if n == 0:
            self._one = vector(k, [])
        else:
            v = vector(k, (n - 1) * ([1] + n * [0]) + [1])
            self._one = B.solve_left(v)

    def _repr_(self):
        return "Finite algebra of degree %d over %s"\
            % (self.degree(), self.base_ring())

    def _coerce_map_from_(self, S):
        return S is self or self.base_ring().has_coerce_map_from(S)

    def _element_constructor_(self, x):
        return FiniteAlgebraElement(self, x)

#    def __nonzero__(self):
#        return bool(self.degree())

    def table(self):
        """
        Return the multiplication table of self.
        """
        return self._table

    def degree(self):
        """
        Return the degree of self over its base field.
        """
        return self._degree

    def _an_element_(self):
        return self.one()

    def basis(self):
        """
        Return a list of FiniteAlgebraElements forming a basis of self.
        """
        return [FiniteAlgebraElement(self, x) for x in self.table()]

    def ideal(self, gens=None, given_by_matrix=False):
        """
        Return the ideal of self generated by gens.

        INPUT:
         - ``gens`` (default: None) - either an element of self or a list
           of elements of self, given as vectors, matrices, or
           FiniteAlgebraElements.  If ``given_by_matrix`` is True, then
           ``gens`` should instead be a matrix whose rows form a basis of
           an ideal of self.
         - ``given_by_matrix`` - bool (default: False)

        OUTPUT:
         - FiniteAlgebraIdeal - the ideal generated by gens.
        """
        return FiniteAlgebraIdeal(self, gens, given_by_matrix)

    def quotient_map(self, ideal):
        """
        Return the quotient of self by ideal.

        INPUT:
         - ``ideal`` - FiniteAlgebraIdeal

        OUTPUT:
         - FiniteAlgebraMorphism
        """
        k = self.base_ring()
        f = ideal.basis_matrix().transpose().kernel().basis_matrix().echelon_form().transpose()
        pivots = f.pivot_rows()
        table = []
        for p in pivots:
            v = vector(k, self.degree())
            v[p] = 1
            v = FiniteAlgebraElement(self, v)
            table.append(f.solve_right(v.matrix() * f))
        B = FiniteAlgebra(k, table)
        return FiniteAlgebraMorphism(self, B, f)

    def maximal_ideal(self):
        """
        Assuming self is a local k-algebra, compute its maximal ideal.

        EXAMPLES::
        
            sage: A = FiniteAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]), Matrix([[0, 1], [0, 0]])])
            sage: A.maximal_ideal()
            Ideal of Finite algebra of degree 2 over Finite Field of size 3 given by basis matrix [0 1]
            sage: B = FiniteAlgebra(QQ, [Matrix([[1,0,0], [0,1,0], [0,0,0]]), Matrix([[0,1,0], [0,0,0], [0,0,0]]), Matrix([[0,0,0], [0,0,0], [0,0,1]])])
            sage: B.maximal_ideal()
            Traceback (most recent call last):
            ...
            ValueError: algebra is not local
        """
        if self.degree() == 0:
            raise ValueError, "the zero algebra is not local"
        gens = []
        for x in self.basis():
            f = x.characteristic_polynomial().factor()
            if len(f) != 1:
                raise ValueError, "algebra is not local"
            if f[0][1] > 1:
                gens.append(f[0][0](x))
        return FiniteAlgebraIdeal(self, gens)

    def primary_decomposition(self):
        """
        Return a list consisting of the quotient maps f: self -> A,
        where A runs over the primary factors of self.

        EXAMPLES::

            sage: A = FiniteAlgebra(GF(3), [Matrix([[1, 0], [0, 1]]), Matrix([[0, 1], [0, 0]])])
            sage: A.primary_decomposition()
            [Morphism from Finite algebra of degree 2 over Finite Field of size 3 to Finite algebra of degree 2 over Finite Field of size 3 given by matrix [1 0]
            [0 1]]
            sage: B = FiniteAlgebra(QQ, [Matrix([[1,0,0], [0,1,0], [0,0,0]]), Matrix([[0,1,0], [0,0,0], [0,0,0]]), Matrix([[0,0,0], [0,0,0], [0,0,1]])])
            sage: B.primary_decomposition()
            [Morphism from Finite algebra of degree 3 over Rational Field to Finite algebra of degree 1 over Rational Field given by matrix [0]
            [0]
            [1], Morphism from Finite algebra of degree 3 over Rational Field to Finite algebra of degree 2 over Rational Field given by matrix [1 0]
            [0 1]
            [0 0]]
        """
        k = self.base_ring()
        n = self.degree()
        if n == 0:
            return []
        # Start with the trivial decomposition of self.
        components = [identity_matrix(k, n)]
        for b in self.table():
            # Use the action of the basis element b to refine our
            # decomposition of self.
            components_new = []
            for c in components:
                # Compute the matrix of b on the component c, find its
                # characteristic polynomial, and factor it.
                b_c = c.solve_left(c * b)
                fact = b_c.characteristic_polynomial().factor()
                if len(fact) == 1:
                    components_new.append(c)
                else:
                    for f in fact:
                        h, a = f
                        e = h(b_c) ^ a
                        ker_e = e.kernel().basis_matrix()
                        components_new.append(ker_e * c)
            components = components_new
        quotients = []
        for i in range(len(components)):
            I = Matrix(k, 0, n)
            for j in range(len(components)):
                if j != i:
                    I = I.stack(components[j])
            quotients.append(self.quotient_map(self.ideal(I, given_by_matrix=True)))
        return quotients

    def maximal_ideals(self):
        """
        Return a list consisting of the maximal ideals of self.
        
        EXAMPLES:
        TODO.
        """
        P = self.primary_decomposition()
        return [f.inverse_image(f.target().maximal_ideal()) for f in P]


class FiniteAlgebraElement(RingElement):
    """
    Create an element of a FiniteAlgebra.

    INPUT:
     - ``A`` - FiniteAlgebra
     - ``elt`` (default: None) - vector or matrix
    
    If ``elt`` is a vector, it is interpreted as a vector of
    coordinates with respect to the given basis of ``A``.  If ``elt``
    is a matrix, it is interpreted as a multiplication matrix with
    respect to this basis.

    EXAMPLES:
    TODO: examples of addition, subtraction, multiplication.
    """

    def __init__(self, A, elt=None):
        RingElement.__init__(self, A)
        n = A.degree()
        if elt is None:
            self._vector = vector(A.base_ring(), n)
            self._matrix = Matrix(A.base_ring(), n)
        else:
            if isinstance(elt,int):
                elt = Integer(elt)
            if A.base_ring().has_coerce_map_from(elt.parent()):
                e = A.base_ring()(elt)
                self._vector = A._one * e
                self._matrix = identity_matrix(A.base_ring(), n) * e
            elif isinstance(elt, sage.matrix.matrix.Matrix):
                self._vector = A._one * elt
                self._matrix = elt
            elif elt.is_vector():
                self._vector = elt
                self._matrix = sum([elt[i] * A.table()[i] for i in range(n)])
            else:
                raise TypeError, "elt should be a matrix, a vector or an element of the base field"

    def _repr_(self):
        return self.matrix().__repr__()

    def __eq__(self, other):
        return self.vector() == self.parent()(other).vector()

    def __ne__(self, other):
        return self.vector() != self.parent()(other).vector()

    def __gt__(self, other):
        raise TypeError

    def __lt__(self, other):
        raise TypeError

    def __ge__(self, other):
        raise TypeError

    def __le__(self, other):
        raise TypeError

    # We have to implement this to make things like 1 == A(1) work,
    # but this also prevents us from throwing an error for 2 > A(1).
    # To make this work, we probably have to use Cython; see
    # sage/structure/element.pyx (class Element).
    def __cmp__(self, other):
        if self.vector() == other.vector():
            return False
        else:
            return True

    def _add_(self, other):
        return FiniteAlgebraElement (self.parent(),
                                     self.vector() + other.vector())

    def _sub_(self, other):
        return FiniteAlgebraElement (self.parent(),
                                     self.vector() - other.vector())

    def _mul_(self, other):
        return FiniteAlgebraElement(self.parent(),
                                    self.vector() * other.matrix())

    def _lmul_(self, other):
        if self.parent().base_ring().has_coerce_map_from(other.parent()):
            return FiniteAlgebraElement(self.parent(),
                                        other * self.vector())
        else:
            raise TypeError, \
                "unsupported operand parent(s) for '*': '%s' and '%s'" \
                % (self.parent(), other.parent())

    def _rmul_(self, other):
        if  self.parent().base_ring().has_coerce_map_from(other.parent()):
            return FiniteAlgebraElement(self.parent(),
                                        other * self.vector())
        else:
            raise TypeError, \
                "unsupported operand parent(s) for '*': '%s' and '%s'" \
                % (self.parent(), other.parent())

    def __pow__(self, n):
        """
        Power by integer.

        EXAMPLES:
        TODO.
        """
        return FiniteAlgebraElement(self.parent(), self.matrix() ^ n)

    def vector(self):
        """
        Return self as a vector.
        
        EXAMPLES: TODO
        """
        return self._vector

    def matrix(self):
        """
        Return self as a matrix.
        
        EXAMPLES: TODO
        """
        return self._matrix

    def minimal_polynomial(self):
        """
        Return the minimal polynomial of self.
        
        EXAMPLES: TODO
        """
        return self.matrix().minimal_polynomial()

    def characteristic_polynomial(self):
        """
        Return the characteristic polynomial of self.
        
        EXAMPLES: TODO
        """
        return self.matrix().characteristic_polynomial()


class FiniteAlgebraIdeal(SageObject):
    """
    Create an ideal of a FiniteAlgebra.

    INPUT:
     - ``A`` - FiniteAlgebra
     - ``gens`` (default: None) - either an element of ``A`` or a list
       of elements of ``A``, given as vectors, matrices, or
       FiniteAlgebraElements.  If ``given_by_matrix`` is True, then
       ``gens`` should instead be a matrix whose rows form a basis of
       an ideal of A.
     - ``given_by_matrix`` - bool (default: False)
    
    EXAMPLES:
    TODO.
    """

    def __init__(self, A, gens=None, given_by_matrix=False):
        self._base_ring = A
        k = A.base_ring()
        n = A.degree()
        if given_by_matrix:
            # TODO: check validity of input
            self._basis_matrix = gens
        elif gens is None:
            self._basis_matrix = Matrix(k, 0, n)
        elif isinstance(gens, list):
            B = [FiniteAlgebraIdeal(A, x).basis_matrix() for x in gens]
            B = reduce(lambda x, y: x.stack(y), B, Matrix(k, 0, n))
            self._basis_matrix = B.echelon_form().image().basis_matrix()
        else:
            if isinstance(gens, sage.matrix.matrix.Matrix):
                gens = FiniteAlgebraElement(gens)
            if isinstance(gens, FiniteAlgebraElement):
                gens = gens.vector()
                B = Matrix([gens * b for b in A.table()])
                self._basis_matrix = B.echelon_form().image().basis_matrix()

    def basis_matrix(self):
        return self._basis_matrix

    def base_ring(self):
        return self._base_ring

    def _repr_(self):
        return "Ideal of %s given by basis matrix %s" \
            % (self.base_ring(), self.basis_matrix())


class FiniteAlgebraMorphism(SageObject):
    """
    Create a morphism between FiniteAlgebras.

    INPUT:
     - ``A`` - FiniteAlgebra
     - ``B`` - FiniteAlgebra
     - ``f`` - matrix of the underlying k-linear map
     - ``check`` (default: True) - check whether the given k-linear
       map really defines a k-algebra homomorphism

    The algebras ``A`` and ``B`` should be defined over the same base
    field k.

    EXAMPLES:
    TODO.
    """

    def __init__(self, A, B, f, check=True):
        self._source = A
        self._target = B
        self._matrix = f
        if check:
            for a in A.basis():
                for b in A.basis():
                    if self(a) * self(b) != self(a * b):
                        raise ValueError, "map is not a ring homomorphism"

    def __call__(self, x):
        x = self.source()(x)
        return FiniteAlgebraElement(self.target(), x.vector() * self.matrix())

    def _repr_(self):
        return "Morphism from %s to %s given by matrix\n%s" \
            % (self.source(), self.target(), self.matrix())

    def source(self):
        return self._source

    def target(self):
        return self._target

    def matrix(self):
        return self._matrix

    def inverse_image(self, I):
        """
        Return the inverse image of ``I`` under self.
        
        INPUT:
        - ``I`` - FiniteAlgebraIdeal, an ideal of self.target()

        OUTPUT:
        - FiniteAlgebraIdeal, the inverse image of ``I`` under self.
        """
        coker_I = I.basis_matrix().transpose().kernel().basis_matrix().transpose()
        return self.source().ideal((self.matrix() * coker_I).kernel().basis_matrix(), given_by_matrix=True)


def bilinear_map_to_algebra(k, mu, tries=Infinity):
    """
    Let k be a field, let A be a finite commutative k-algebra, let F
    and F' be free A-modules of rank 1, and let mu: F x F -> F' be a
    perfect A-bilinear map, i.e. one that induces an isomorphism from
    F \tensor_A F to F'.  Given the matrices of the maps F -> F'
    sending g to mu(g, f), for f ranging over a basis of F and with
    respect to the same basis of F and any basis of F', this function
    computes A as a FiniteAlgebra such that the action of A on F is
    given by standard matrix multiplication.

    INPUT:
     - ``k`` - the base field
     - ``mu`` - list of n matrices of size n x n over a field k
       defining a bilinear map as above.
     - ``tries`` - Integer or Infinity (default: Infinity) - maximal
       number of tries to find a generator of F.

    OUTPUT:
     - FiniteAlgebra

    EXAMPLES::

        sage: bilinear_map_to_algebra(QQ, [])
        Finite algebra of degree 0 over Rational Field
        sage: bilinear_map_to_algebra(QQ, [Matrix([5])])
        Finite algebra of degree 1 over Rational Field
    """
    n = len(mu)
    V = k^n
    while tries > 0:
        tries -= 1
        v = V.random_element()
        f = sum([v[i] * mu[i] for i in range(n)], Matrix(k, n))
        if f.rank() < n:
            continue
        finv = f.inverse()
        return FiniteAlgebra (k, [g * finv for g in mu])
    raise ArithmeticError, "cannot find a generator"
