# Let us postpone Parent/Element structures for the moment, until we
# really want to make use of coercion.

if 'ATTACH' not in globals():
    from addflip_chain import chain_with_computation, addflip_chain
    from bundled_curve import BundledCurve, BundledCurveEffectiveDivisor
    

class JacobianModel(SageObject):
    """
    Jacobian of a bundled curve given by a projective model of the curve.

    INPUT:
     - curve - BundledCurve
     - model - "small", "medium" or "large" (default: "medium")

    EXAMPLES::
        sage: C = BundledModularCurve(GF(3), Gamma1(13), 2, 9)
        sage: J = JacobianModel(C)
        sage: P = J.random_element()
        sage: J.order() * P == 0
        True
        sage: P in J
        True
        (-2) * P in J
        sage: 3 * P - 4 * P == -P
        True

    TODO:
     - implement small and large model as well
     - document methods
    """

    def __init__(self, curve, model="medium"):
        if curve.genus() == 0:
            raise ValueError, "genus of curve must be positive"
        if model != "medium":
            raise NotImplementedError, "only medium model has been implemented"
        self._curve = curve
        self._model = model

    def base_extend(self, K):
        """
        Return self, base changed to K.

        INPUT:
         - ``K`` - field

        OUTPUT: JacobianModel
        """
        result = JacobianModel(self._curve.base_extend(K), model=self._model)
        # TODO: adapt zero element
        return result

    def curve(self):
        return self._curve

    def model(self):
        return self._model

    def base_ring(self):
        return self.curve().base_ring()

    def _repr_(self):
        return "Jacobian of %s, %s model" % (self.curve(), self.model())

    def __contains__(self, x):
        if self is not x.parent():
            return False
        C = self.curve()
        D = x.divisor()
        if self.model() == "medium":
            return (D + C.zero_divisor(3)).degree() == C.degree()
        else:
            raise NotImplementedError, "only medium model has been implemented"

    def zero_element(self):
        try:
            return self._zero_element
        except AttributeError:
            pass
        C = self.curve()
        if self.model() == "medium":
            s = C.sections(1)[0]
            V = C.multiplication_matrix(1, s, 1)
            D = BundledCurveEffectiveDivisor(C, 2, V)
            self._zero_element = JacobianElement(self, D)
        else:
            raise NotImplementedError, "only medium model has been implemented"
        return self._zero_element

    def random_element(self, tries=Infinity):
        C = self.curve()
        if self.model() == "medium":
            D = C.random_divisor(C.degree(), tries=tries).simplify(2)
            return JacobianElement(self, D)
        else:
            raise NotImplementedError, "only medium model has been implemented"

    def order(self):
        try:
            return self.curve().zeta_polynomial()(1)
        except (ValueError, AttributeError):
            raise ArithmeticError, "no method known to compute the order"


class JacobianElement(SageObject):
    """
    Create an element of a Jacobian.

    INPUT:
     - jacobian - JacobianModel
     - divisor - BundledCurveEffectiveDivisor
    """

    def __init__(self, jacobian, divisor):
        self._parent = jacobian
        self._divisor = divisor

    def _repr_(self):
        return "Point on %s" % self.parent()

    def parent(self):
        return self._parent

    def divisor(self):
        return self._divisor

    def model(self):
        return self.parent().model()

    def curve(self):
        return self.parent().curve()

    def _addflip(self, other):
        """
        Return -(self + other).
        """
        if self.parent() is not other.parent():
            raise TypeError, "addflip requires parents to be the same"
        C = self.curve()
        if self.model() == "medium":
            D_plus_E = (self.divisor() + other.divisor()).simplify(3)
            s = D_plus_E.basis()[0]
            V_DEF = C.multiplication_matrix(3, s, 2)
            V_F = C.divide_subspaces(V_DEF, 5, D_plus_E.basis(), 3)
            F = BundledCurveEffectiveDivisor(C, 2, V_F.basis_matrix())
            return JacobianElement(self.parent(), F)
        else:
            raise NotImplementedError, "only medium model has been implemented"

    def __add__(self, other):
        return -self._addflip(other)

    def __sub__(self, other):
        return other._addflip(-self)

    def __neg__(self):
        if self.model() == "medium":
            C = self.curve()
            D = self.divisor()
            s = D.basis()[0]
            V_DE = C.multiplication_matrix(2, s, 2)
            V_E = C.divide_subspaces(V_DE, 4, D.basis(), 2)
            E = BundledCurveEffectiveDivisor(C, 2, V_E.basis_matrix())
            return JacobianElement(self.parent(), E)
        else:
            raise NotImplementedError, "only medium model has been implemented"

    def __rmul__(self, scalar):
        multiples = {0: self.parent().zero_element(), 1: self}
        chain = chain_with_computation(addflip_chain(scalar))
        for x, y, z in chain[2:]:
            # print "Computing %s * self" % z
            if x == 0 or y == 0:
                multiples[z] = -multiples[-z]
            else:
                multiples[z] = multiples[x]._addflip(multiples[y])
        return multiples[scalar]

    def __nonzero__(self):
        C = self.curve()
        if self.model() == "medium":
            V = C.divide_subspaces(self.divisor().basis_matrix(), 2,
                                   C.sections_igs(1), 1)
            return V.dimension() == 0
        else:
            raise NotImplementedError, "only medium model has been implemented"

    def __eq__(self, other):
        try:
            return not bool(self - other)
        except AttributeError:
            if other == 0:
                return not bool(self)
            else:
               raise

    def __ne__(self, other):
        return not(self == other)

    def __gt__(self, other):
        raise TypeError

    def __lt__(self, other):
        raise TypeError

    def __ge__(self, other):
        raise TypeError

    def __le__(self, other):
        raise TypeError

    def frobenius(self, n=1):
        """
        Apply the absolute Frobenius Frob_p^n to self.

        INPUT:
         - ``n`` - integer (default: 1)

        OUTPUT: JacobianElement

        EXAMPLES::
            sage: C = BundledModularCurve(GF(3^2, 'a'), Gamma1(13), 2, 9)
            sage: J = JacobianModel(C)
            sage: P = J.random_element()
            sage: Q = P.frobenius()
            sage: P == Q.frobenius()
            True
            sage: 19 * (P + Q) == 0  # 19 is #J(F_3)
            True
        """
        return JacobianElement(self.parent(), self.divisor().frobenius(n))

