if 'ATTACH' not in globals():
    from bundled_curve import BundledCurve, BundledCurveEffectiveDivisor


# The following functions are workarounds for NotImplemented stuff in Sage.

def _hecke_operator(f, n):
    """
    f is a modular form over F_p
    """
    MQ = ModularForms(f.group(), f.weight())
    MFp = f.parent()
    Fp = MFp.base_ring()
    val = MFp.echelon_basis()[-1].valuation() + 1
    mat = matrix([vector(g.coefficients(range(val))) for g in MQ.integral_basis()])
    v = vector(f.coefficients(range(val)))
    w = mat.change_ring(Fp).solve_left(v).change_ring(ZZ)
    fZZ = sum([w[i] * MQ.integral_basis()[i] for i in range(len(w))])
    return MFp(MQ.T(n)(fZZ).q_expansion(MQ.sturm_bound()))


def hecke_matrix(M, n):
    """
    Determine the matrix of T(n) relative to M.basis().

    INPUT:
     - ``M`` - space of modular forms over field of positive characteristic.
     - ``n`` - integer

    OUTPUT: matrix
    """
    K = M.base_ring()
    p = K.characteristic()
    if p == 0:
        return M.hecke_matrix(n)
    return matrix([M.coordinate_vector(_hecke_operator(f, n)) for
                   f in M.basis()])


def diamond_matrix(M, d):
    """
    Determine the matrix of <d> relative to M.basis().

    INPUT:
     - ``M`` - space of modular forms
     - ``d`` - integer

    OUTPUT: matrix

    TODO: this should be made much faster!
    """
    result = hecke_matrix(M, 1)
    N = M.level()
    k = M.weight()
    L = M.base_ring().characteristic()
    if gcd(d, N) != 1:
        return 0 * result
    d = d % N
    if L.divides(d):
        d = d + N
    for p, e in d.factor():
        # T(p^2) = T(p)^2 - p^(k-1)<p>
        Tp2 = hecke_matrix(M, p^2)
        Tp = hecke_matrix(M, p)
        result *= ((Tp^2 - Tp2) * GF(L)(p)^(1-k))^e
    return result


class BundledModularCurve(BundledCurve):
    """
    Create the modular curve X_G with line bundle omega^k(-r[oo]) on it.

    INPUT:
     - ``K`` - field (base field of the curve) or ZZ
     - ``G`` - group containing Gamma_1(N) for some N
     - ``k`` - integer
     - ``r`` - integer

    EXAMPLES::
        sage: C = BundledModularCurve(GF(3), Gamma1(13), 2, 9)
        sage: C.degree()
        5
        sage: C.genus()
        2
        sage: D = C.random_divisor(5)
        sage: D.degree()
        5
        sage: E = C.random_divisor(4)
        sage: E.degree()
        4
        sage: F = D + E
        sage: F.degree()
        9
        sage: F.i()
        4
        sage: F.simplify().i()
        3

    REMARK:
     - Because of the choice of line bundle, we assume the cusp oo at
       infinity is rational, i.e. we are using the model for X_1(N) that
       parametrises pairs (E, i) with E a generalised elliptic curve and
       i: mu_N -> E a closed immersion.

    TODO:
     - more examples
     - make everything work for G that contains Atkin-Lehner operators
    """

    def __init__(self, K, G, k, r):
        if Integer(K.characteristic()).divides(G.level()):
            raise ValueError, "curve is singular"
        if K is ZZ:
            M_k = ModularForms(G, k, QQ)
        else:
            M_k = ModularForms(G, k, K)
        basis_M_k = M_k.echelon_basis()
        self._max_power_of_L = 7  # This suffices for our purposes
        q_bound = self._max_power_of_L * basis_M_k[-1].valuation() + 1
        pol_ring.<q> = K[]
        ring = pol_ring.quo(q^q_bound)
        if K is ZZ:
            basis = [f for f in M_k.q_integral_basis(q_bound) if f.valuation() >= r]
            sections = [ring(list(f)) for f in basis]
        else:
            basis = [f for f in basis_M_k if f.valuation() >= r]
            sections = [ring(list(f.q_expansion(q_bound))) for f in basis]
        BundledCurve.__init__(self, ring, sections)
        self._group = G
        self._k = k
        self._r = r

    def base_extend(self, K):
        """
        Return self, base changed to K.

        INPUT:
         - ``K`` - field

        OUTPUT: BundledModularCurve

        EXAMPLES: TODO
        """
        result = deepcopy(self)
        result._base_field = K
        q_bound = self._ring.modulus().degree()
        pol_ring.<q> = K[]
        ring = pol_ring.quo(q^q_bound)
        result._ring = ring
        for i in range(len(self._V)):
            result._V[i] = [ring(x.list()) for x in self._V[i]]
            rv = self._V_ring_vectors[i].base_extend(K)
            result._V_ring_vectors[i] = rv
        for i in self._igs:
            result._igs[i] = [ring(x.list()) for x in self._igs[i]]
        return result

    def group(self):
        return self._group

    def level(self):
        return self.group().level()

    def weight(self):
        """
        Return the modular forms' weight defining the line bundle sections.
        """
        return self._k

    def modular_forms(self, i):
        """
        Return the ambient modular forms space associated to H^0(L^i).

        INPUT:
         - ``i`` - integer

        OUTPUT: ModularForms
        """
        return ModularForms(self.group(), i * self.weight(), self.base_ring())

    def _repr_(self):
        if self._r == 0:
            lb = "line bundle omega^%s" % self._k
        else:
            lb = "line bundle omega^%s(-%s[oo])" % (self._k, self._r)
        s = "Modular curve X_G over %s with G = %s and line bundle %s"
        return "Modular curve X_G over %s with G = %s and %s" %\
               (self.base_ring(), self.group(), lb)

    def _latex_curve(self):
        from sage.modular.arithgroup import congroup_gamma0, congroup_gamma1
        if isinstance(self.group(), congroup_gamma0.Gamma0_class):
            name = "X_0(%d)" % self.level()
        elif isinstance(self.group(), congroup_gamma1.Gamma1_class):
            name = "X_1(%d)" % self.level()
        else:
            grp = str(self.group()).replace('Congruence Subgroup ', '')
            name = "X(%s)" % grp
        name = name.replace('Gamma', '\\Gamma')
        return name + "_{%s}" % latex(self.base_ring())
    
    def _latex_bundle(self):
        lb = "\\omega^{\\otimes %s}" % self._k
        if self._r == 0:
            return lb
        else:
            return lb + "(-%s\\cdot\\infty)" % self._r
    
    def _latex_(self):
        return "\\left(%s, %s\\right)" % (self._latex_curve(), self._latex_bundle())
    
    def distinguished_point(self):
        """
        Return the rational point defined by the cusp oo.

        OUTPUT: BundledCurveEffectiveDivisor

        EXAMPLES: TODO
        """
        return BundledCurveEffectiveDivisor(self, 1, self.sections(1)[1:])

    def zeta_polynomial(self):
        """
        If self is defined over F_q, return the numerator of the zeta function.

        OUTPUT: polynomial with integer coefficients

        EXAMPLES::
            sage: C = BundledModularCurve(GF(5), Gamma1(17), 2, 13)
            sage: C.zeta_polynomial()
            3125*T^10 + 1250*T^9 + 875*T^8 - 300*T^7 - 100*T^6 - 156*T^5 - 20*T^4 - 12*T^3 + 7*T^2 + 2*T + 1
        """
        K = self.base_ring()
        if not K.is_finite():
            raise ValueError, "a curve over an infinite field has no zeta function"
        p = K.characteristic()
        g = self.genus()
        A = MatrixSpace(QQ, g)(0)
        B = MatrixSpace(QQ, g)(1)
        S = CuspForms(self.group(), 2)
        C = (S.T(p^2) - S.T(p)^2).matrix()
        D = S.T(p).matrix()
        M = block_matrix([[A, B], [C, D]]) ^ K.degree()
        return ZZ['T'](M.charpoly().reverse())

    def diamond_matrix(self, i, n):
        """
        Return the matrix for <n> on H^0(L^i), if L is a power of omega.

        INPUT:
         - ``i`` - integer
         - ``n`` - integer

        OUTPUT: matrix
        """
        if self._r != 0:
            raise ValueError, "line bundle must be power of omega"
        N = self.level()
        n = n % N
        try:
            return self._diamond_matrix[(i, n)]
        except AttributeError:
            self._diamond_matrix = {}
        except KeyError:
            pass
        M = self.modular_forms(i)
        K = self.base_ring()
        P = PowerSeriesRing(K, 'q')
        power_series_basis = [P(s.list()).add_bigoh(M.sturm_bound()) for
                              s in self.sections(i)]
        basis_matrix = matrix([M.coordinate_vector(M(f)) for
                               f in power_series_basis])
        output_matrix = basis_matrix * diamond_matrix(M, n)
        # resulting_matrix * basis_matrix = output_matrix
        self._diamond_matrix[(i, n)] = basis_matrix.solve_left(output_matrix)
        return self._diamond_matrix[(i, n)]
