if 'ATTACH' not in globals():
    from bundled_curve import BundledCurve, BundledCurveEffectiveDivisor
    from modular_curve import BundledModularCurve
    from jacobian import JacobianModel, JacobianElement

def ModularJacobianModel(K, G, model="medium", omega_power=False):
    """
    Return a model for the Jacobian of the modular curve X(G) over K.

    INPUT:
     - ``K`` - field
     - ``G`` - group containing Gamma1(N) for some N
     - ``model`` - "small", "medium" or "large" (default: "medium")
     - ``omega_power`` - bool (default: False)

    OUTPUT: JacobianModel

    REMARKS:
     - If omega_power is True, a line bundle that is a power of omega
       will be chosen.  If it is False, then a line bundle omega^k(-r[oo])
       of minimal degree will be chosen.

    TODO:
     - make everything work for G that contains Atkin-Lehner operators
     - examples
    """
    g = G.genus()
    if g == 0:
        raise ValueError, "genus must be positive"
    # Compute the least k0 > 0 for which omega^k0 is a line bundle
    # on the (coarse) modular curve.  The following lines take into
    # account all possibilities for the orders of non-trivial
    # stabilisers (2, 3, 4 or 6), at least in char. not 2 or 3.
    # Note that G.nu2() > 0 only if -1 is in G, and the result
    # of this function is then the number of points where the
    # elliptic curve with G-structure has an automorphism of order 4.
    k0_divisors = []
    if G.is_even() or G.nirregcusps() > 0:
        k0_divisors.append(2)
    if G.nu2() > 0:
        k0_divisors.append(4)
    if G.nu3() > 0:
        k0_divisors.append(3)
    k0 = lcm(k0_divisors)
    k = max(2, k0 * ceil((2*g + 1) * 12 / (G.projective_index() * k0)))
    d = k * G.projective_index() // 12
    if omega_power == True:
        r = 0
    else:
        r = d - (2*g + 1)
    C = BundledModularCurve(K, G, k, r)
    return JacobianModel(C, model=model)


class ModularJacobian(SageObject):
    """
    Jacobian of modular curve X(G) over K.

    INPUT:
     - ``K`` - field
     - ``G`` - group containing Gamma1(N) for some N

    The created object will contain several JacobianModels.

    EXAMPLES: TODO
    """

    def __init__(self, K, G):
        self._jac_medpow = \
             ModularJacobianModel(K, G, model="medium", omega_power=True)
        self._jac_medmin = \
             ModularJacobianModel(K, G, model="medium", omega_power=False)
        self._r = self.jac("medmin").curve()._r
        self._group = G
        self._base_ring = K
        self._pref_model = "medmin"

    def base_extend(self, K):
        """
        Return self base changed to K.

        INPUT:
         - ``K`` - field

        OUTPUT: ModularJacobian

        EXAMPLES::
            sage: J = ModularJacobian(GF(5), Gamma1(13))
            sage: J2 = J.base_extend(GF(5^3, 'a'))
            sage: P2 = J2.random_element()
            sage: 2^6 * 13 * 19 * P2 == 0
            True
            sage: P = J.random_element()
            sage: 19 * P == 0
            True
            sage: J3 = J.base_extend(GF(5^3, 'a'))
            sage: P3 = J3.random_element()
            sage: 2^6 * 13 * 19 * P3 == 0
            True
        """
        result = copy(self)
        result._base_ring = K
        result._jac_medpow = self._jac_medpow.base_extend(K)
        result._jac_medmin = self._jac_medmin.base_extend(K)
        return result

    def _repr_(self):
        return "Jacobian of modular curve X_G with G = %s over %s" %\
               (self.group(), self.base_ring())

    def _latex_(self):
        from sage.modular.arithgroup import congroup_gamma0, congroup_gamma1
        if isinstance(self.group(), congroup_gamma0.Gamma0_class):
            name = "J_0(%d)" % self.level()
        elif isinstance(self.group(), congroup_gamma1.Gamma1_class):
            name = "J_1(%d)" % self.level()
        else:
            grp = str(self.group()).replace('Congruence Subgroup ', '')
            name = "J(%s)" % grp
        name = name.replace('Gamma', '\\Gamma')
        return name + "_{%s}" % latex(self.base_ring())
        
    def base_ring(self):
        return self._base_ring

    def group(self):
        return self._group

    def level(self):
        return self.group().level()

    def pref_model(self):
        """
        Return the preferred model of self.

        OUTPUT: string

        EXAMPLES: TODO
        """
        return self._pref_model

    def jac(self, model):
        """
        Return the JacobianModel attached to "model".

        INPUT:
         - ``model`` - "medmin" or "medpow" (to be expanded)

        OUTPUT: JacobianModel

        EXAMPLES: TODO
        """
        if model == "medmin":
            return self._jac_medmin
        elif model == "medpow":
            return self._jac_medpow
        raise NotImplementedError, "requested model has not been implemented"

    def curve(self, model=None):
        """
        Return a bundled modular curve that self is the Jacobian of.

        INPUT:
         - ``model`` - None (the default) "medmin" or "medpow" for now

        OUTPUT: BundledModularCurve

        EXAMPLES: TODO
        """
        if model is None:
            model = self.pref_model()
        return self.jac(model).curve()

    def random_element(self, tries=Infinity):
        model = self.pref_model()
        return ModularJacobianElement(self, model,
                                 self.jac(model).random_element(tries=tries))

    def order(self):
        return self.jac(self.pref_model()).order()

    def zero_element(self):
        model = self.pref_model()
        return ModularJacobianElement(self, model,
                                      self.jac(model).zero_element())

    def __contains__(self, point):
        if self is not point.parent():
            return False
        model = self.pref_model()
        return point.point(model) in self.jac(model)

    def __call__(self, point):
        """
        If self is J.base_extend(bla) and point is in J, extend point to self.

        INPUT:
         - ``point`` - ModularJacobianElement

        OUTPUT: ModularJacobianElement

        EXAMPLES::
            sage: J = ModularJacobian(GF(5), Gamma1(13))
            sage: J2 = J.base_extend(GF(5^3, 'a'))
            sage: P = J.random_element()
            sage: P2 = J2(P)
            sage: P2 in J2
            True
            sage: 19 * P2 == 0
            True
            sage: Q2 = J2.random_element()
            sage: P2 + Q2 in J2
            True
        """
        model = self.pref_model()
        jac = self.jac(model)
        C = jac.curve()
        D = C(point.point(model).divisor())
        return ModularJacobianElement(self, model, JacobianElement(jac, D))


class ModularJacobianElement(SageObject):
    """
    Create an element of a modular Jacobian.

    INPUT:
     - ``jacobian`` - JacobianModel
     - ``model`` - for now "medmin" or "medpow"
     - ``point`` - JacobianElement
    """
    def __init__(self, jacobian, model, point):
        self._parent = jacobian
        if model == "medpow":
            self._point_medpow = point
        elif model == "medmin":
            self._point_medmin = point
        else:
            raise NotImplementedError, "model must be medpow or medmin"

    def _repr_(self):
        return "Point on %s" % self.parent()

    def parent(self):
        return self._parent

    def base_ring(self):
        return self.parent().base_ring()

    def pref_model(self):
        return self.parent().pref_model()

    def point_medpow(self):
        try:
            return self._point_medpow
        except AttributeError:
            pass
        if hasattr(self, "_point_medmin"):
            self._convert_medmin_to_medpow()
            return self._point_medpow
        raise ArithmeticError, "cannot compute model for point"

    def point_medmin(self):
        try:
            return self._point_medmin
        except AttributeError:
            pass
        if hasattr(self, "_point_medpow"):
            self._convert_medpow_to_medmin()
            return self._point_medmin
        raise ArithmeticError, "cannot compute model for point"

    def point(self, model):
        if model == "medpow":
            return self.point_medpow()
        elif model == "medmin":
            return self.point_medmin()
        raise NotImplementedError, "requested model has not been implemented"

    def _convert_medmin_to_medpow(self):
        """
        Given self._point_medmin, compute and store self._point_medpow.
        """
        # Let L1 be the line bundle of medmin and L2 that of medpow.
        # Then L2 = L1(rP).  The point is given as H^0(L1^2(-D)).
        # First step: compute H^0(L1^2(-D)) as a subspace of H^0(L2^2)
        D = self._point_medmin.divisor()
        J2 = self.parent().jac("medpow")
        C2 = J2.curve()
        D2 = BundledCurveEffectiveDivisor(C2, 2, D.basis())
        # D2 is the divisor D + 2rP, the point in the Jacobian "is" D + rP
        # This means we have to divide by rP, which is H^0(L1) in H^0(L2)
        V = D2.simplify(3).basis_matrix()
        S = self.parent().jac("medmin").curve().sections(1)
        V2 = C2.divide_subspaces(V, 3, S, 1)
        E = BundledCurveEffectiveDivisor(C2, 2, V2.basis_matrix())
        self._point_medpow = JacobianElement(J2, E)

    def _convert_medpow_to_medmin(self):
        """
        Given self._point_medpow, compute and store self._point_medmin.
        """
        # Let L1 be the line bundle of medmin and L2 that of medpow.
        # Then L2 = L1(rP).  The point is given as H^0(L2^2(-D)).
        # We must find an effective E with E + rP ~ D and compute E + rP.
        # First we compute D' with L2^2(-D-D') trivial.
        Dprime = (-self._point_medpow).divisor()
        # Then we compute s in H^0(L2^2(-D'-rP))
        C2 = self.parent().jac("medpow").curve()
        J1 = self.parent().jac("medmin")
        C1 = J1.curve()
        rP = BundledCurveEffectiveDivisor(C2, 1, C1.sections(1))
        Dprime_rP = (Dprime + rP).simplify(2)
        s = Dprime_rP.basis()[0]
        # Now (s)_L2^2 = D' + rP + E.
        # So E + rP "=" s * H^0(L2^2) / H^0(L2^2(-D'))
        Vs = C2.multiplication_matrix(2, s, 2)
        V_ErP = C2.divide_subspaces(Vs, 4, Dprime.basis(), 2)
        ErP = BundledCurveEffectiveDivisor(C2, 2, V_ErP.basis_matrix())
        # We have H^0(L2^2(-E-rP)) but we want H^0(L1^2(-E)).  We compute
        # H^0(L2^2(-E-2rP)), which is H^0(L1^2(-E)) as subspace of H^0(L1^2).
        E2rP = (ErP + rP).simplify(2)
        E1 = BundledCurveEffectiveDivisor(C1, 2, E2rP.basis())
        self._point_medmin = JacobianElement(J1, E1)

    def diamond(self, d):
        """
        Return <d>(self).

        INPUT:
         - ``d`` - integer

        OUTPUT: ModularJacobianElement

        EXAMPLES::
            sage: J = ModularJacobian(GF(5^2, 'a'), Gamma1(13))
            sage: P = J.random_element()
            sage: P.diamond(-1) == P
            True
            sage: P.diamond(5) in J
            True
            sage: P.diamond(2) in J
            True
            sage: J = ModularJacobian(GF(5^2,'a'), Gamma0(19))
            sage: P = J.random_element()
            sage: P.diamond(2) == P
            True
        """
        if d.gcd(self.parent().level()) != 1:
            raise ValueError, "input must be coprime to level"
        P = self.point("medpow")
        J = P.parent()
        C = J.curve()
        D = P.divisor()
        i = D.i()
        diamond_D = BundledCurveEffectiveDivisor(C, i,
                                   D.basis_matrix() * C.diamond_matrix(i, d))
        diamond_P = JacobianElement(J, diamond_D)
        return ModularJacobianElement(self.parent(), "medpow", diamond_P)

    def frobenius(self, n=1):
        """
        Apply the absolute Frobenius Frob_p^n to self.

        INPUT:
         - ``n`` - integer (default: 1)

        OUTPUT: ModularJacobianElement

        EXAMPLES::
            sage: J = ModularJacobian(GF(7^2, 'a'), Gamma1(15))
            sage: P = J.random_element()
            sage: P.frobenius(2) == P
            True
        """
        model = self.pref_model()
        return ModularJacobianElement(self.parent(), model,
                                      self.point(model).frobenius(n))

    def hecke_at_p(self):
        """
        Return T_p(self) where p is the characteristic of the base field.

        OUTPUT: ModularJacobianElement

        REMARK: we use the "mu"-model, so that T_p = <p>Frob_p + Ver_p.
        """
        p = self.base_ring().characteristic()
        d = self.base_ring().degree()
        P1 = self.diamond(p).point("medpow")
        P2 = (p * self).point("medpow")
        return ModularJacobianElement(self.parent(), "medpow",
                                      P1.frobenius() + P2.frobenius(d - 1))

    def hecke(self, n):
        """
        Apply T(n) to self.

        INPUT:
         - ``n`` - integer

        OUTPUT: ModularJacobianElement

        EXAMPLES::
            sage: J = ModularJacobian(GF(7), Gamma1(15))
            sage: P = J.random_element()
            sage: P.hecke(7) == 0
            True
            sage: P.hecke(49) == P
            True

        REMARK: for now, ``n`` must be a power of char(self.base_ring())
        """
        if n <= 0:
            raise ValueError, "n must be positive"
        factors = n.factor()
        result = self
        for p, e in factors:
            if p != self.base_ring().characteristic():
                raise NotImplementedError, "n must be a power of %s" %\
                                           self.base_ring().characteristic()
            # Use the recursion T(p^(r+1)) = T_p T_(p^r) - <p> p T(p^(r-1))
            prev_result = self.parent().zero_element()
            for r in range(e):
                if r == 0:
                    prev_result, result = result, result.hecke_at_p()
                else:
                    prev_result, result = result, \
                           result.hecke_at_p() - (p * prev_result).diamond(p)
        return result

    def hecke_poly(self, n, f):
        """
        Apply f(T(n)) to self.

        INPUT:
         - ``n`` - integer
         - ``f`` - Polynomial

        OUTPUT: ModularJacobianElement

        EXAMPLES: TODO

        REMARK: for now, ``n`` must be a power of char(self.base_ring())
        """
        d = f.degree()
        T = [self]
        for i in range(d):
            T.append(T[i].hecke(n))
        return sum([f[i] * T[i] for i in range(d)], T[d])

    def normalised_divisor(self):
        """
        Return the O-normalised divisor attached to self,
        where O is the distinguished point on the curve.

        OUTPUT: BundledCurveEffectiveDivisor
        """
        D = self.point(self.pref_model()).divisor()
        return D.normalise()

    def __add__(self, other):
        model = self.pref_model()
        return ModularJacobianElement(self.parent(), model,
                                      self.point(model) + other.point(model))

    def __sub__(self, other):
        model = self.pref_model()
        return ModularJacobianElement(self.parent(), model,
                                      self.point(model) - other.point(model))

    def __neg__(self):
        model = self.pref_model()
        return ModularJacobianElement(self.parent(), model,
                                      -self.point(model))

    def __rmul__(self, scalar):
        model = self.pref_model()
        return ModularJacobianElement(self.parent(), model,
                                   scalar * self.point(model))

    def __nonzero__(self):
        return bool(self.point(self.pref_model()))

    def __eq__(self, other):
        try:
            return not bool(self - other)
        except AttributeError:
            if other == 0:
                return not bool(self)
            else:
                raise

    def __ne__(self, other):
        return not(self == other)

    def __gt__(self, other):
        raise TypeError

    def __lt__(self, other):
        raise TypeError

    def __ge__(self, other):
        raise TypeError

    def __le__(self, other):
        raise TypeError
