class TauRepresentation_local(SageObject):
    """
    Class for computing mod p polynomials for mod L representations for tau.
    
    INPUT:
     - ``rep`` - TauRepresentation
     - ``p`` - prime number, not allowed to be equal to rep.L()
    """
    
    def __init__(self, rep, p):
        self._rep = rep
        self._p = p
        self._eps_p = rep.character(p)  # includes p^(k -1)
        self._ap = rep.coefficient(p)
        R.<t> = rep.coefficient_field()[]
        self._charpoly = t^2 - self._ap * t + self._eps_p
    
    def _repr_(self):
        return "Local mod %s representation for tau at %s" % (self.L(), self.p())
        
    def _latex_(self):
        return latex(self.rep()) + "|_{D_{%s}}" % self.p()
    
    def L(self):
        return self.rep().L()
    
    def rep(self):
        return self._rep
    
    def p(self):
        return self._p
    
    def charpoly(self):
        """
        Return the characteristic polynomial of Frobenius.
        """
        return self._charpoly
    
    def base_field(self):
        # Determine a (not necessarily minimal) field extension
        # of F_p over which the points in the representation space
        # are rational.
        try:
            return self._base_field
        except AttributeError:
            # Simple-minded approach to finding the multiplicative order
            # of t modulo P (since tbar.multiplicative_order() raises
            # NotImplementedError).
            P = self.charpoly()
            Q.<tbar> = P.parent().quotient(P)
            order = 1
            while tbar^order != 1:
                order += 1
            self._base_field = GF(self.p()^order, 'a')
        return self._base_field
    
    def jacobian(self):
        try:
            return self._jacobian
        except AttributeError:
            self._jacobian = self.rep().jacobian().base_extend(GF(self.p()))
        return self._jacobian
    
    def jacobian_base_ext(self):
        try:
            return self._jacobian_base_ext
        except AttributeError:
            self._jacobian_base_ext = self.jacobian().base_extend(self.base_field())
        return self._jacobian_base_ext
    
    def random_point(self, tries=Infinity):
        """
        Return a random non-zero point in the representation space.
        
        INPUT:
         - ``tries`` - Integer or Infinity (default: Infinity)

        OUTPUT: ModularJacobianElement
        
        REMARK:
         - The distribution is not always close to uniform.
        """
        J = self.jacobian_base_ext()
        n = J.order()
        L = self.rep().L()
        if n % L != 0:
            raise ValueError, "Jacobian has no points of order %s" % L
        # Compute the characteristic polynomial of T_p
        # and factor it modulo L as (t - a_p)^o * g, where
        # g is coprime to t - a_p.
        R.<t> = GF(L)[]
        f = R(CuspForms(self.jacobian().group(), 2).T(self.p()).charpoly())
        f0 = t - self._ap
        o = f.ord(f0)
        if o == 0:
            raise ValueError, "Jacobian has no points annihilated by %s" % f0
        while tries > 0:
            tries -= 1
            P = J.random_element()
            print "Random point of Jacobian computed"
            P = n.prime_to_m_part(L) * P
            # Now P is an L-primary torsion point of the Jacobian.
            if P == 0:
                continue
            while True:
                Q = L * P
                if Q == 0:
                    break
                P = Q
            # Now P is a more-or-less random L-torsion point of the Jacobian.
            # Project P onto the subspace on which f0 acts nilpotently.
            P = P.hecke_poly(self.p(), f // f0^o)
            if P != 0:
                return P
        
    def basis(self, tries=Infinity):
        """
        Compute a basis for the representation space
        
        INPUT:
         - ``tries`` - Integer or Infinity (default: Infinity)

        OUTPUT: a tuple of two ModularJacobianElements
        """
        try:
            return self._basis
        except AttributeError:
            P = self.random_point(tries=tries)
            print "Random point of V_%s computed" % self.L()
            multiples_P = [self.jacobian_base_ext().zero_element()]
            for i in range(self.rep().L() - 1):
                multiples_P.append(multiples_P[i] + P)
            # Look for a linearly independent point, using
            # Frob(P) as a first guess.
            Q = P.frobenius()
            print "Trying if Frob(P) is linearly independent of P"
            tries2 = tries
            while tries2 > 0:
                tries2 -= 1
                Q_multiple_of_P = False
                for i in range(self.rep().L()):
                    if Q == multiples_P[i]:
                        break
                if not Q_multiple_of_P:
                    break
                Q = self.random_point(tries=tries)
                print "Another random point of V_%s computed" % self.L()
            self._basis = [P, Q]
        print "Basis for V_%s computed" % self.L()
        return self._basis
    
    def all_points(self, tries=Infinity):
        """
        Compute the set of all points in the representation space.
        
        INPUT:
         - ``tries`` - Integer or Infinity (default: Infinity)
        
        OUTPUT: a dictionary consisting of key-value pairs {v: P} with 
         - ``v`` - element of GF(L)^2
         - ``P`` - ModularJacobianElement
        
        REMARK:
         - The vector space structure corresponds to self.basis().
        """
        L = self.rep().L()
        V = GF(L)^2
        result = {}
        P = self.jacobian_base_ext().zero_element()
        for m in GF(L):
            Q = P
            for n in GF(L):
                # Convert the vector to a tuple because dictionary indices
                # must be hashable.
                result[tuple(V([m, n]))] = Q
                if n != L-1:
                    Q += self.basis(tries=tries)[1]
            if m != L-1:
                P += self.basis()[0]
        print "All points of V_%s computed" % self.L()
        return result
    
    def all_divisors(self, tries=Infinity):
        """
        Compute the set of all normalised divisors corresponding to
        the points in the representation space.
        
        INPUT:
         - ``tries`` - Integer or Infinity (default: Infinity)
        
        OUTPUT: a dictionary consisting of key-value pairs {v: D} with 
         - ``v`` - element of GF(L)^2
         - ``D`` - BundledCurveEffectiveDivisor
        
        REMARK:
         - The vector space structure corresponds to self.basis().
        """
        points = self.all_points(tries=tries);
        # It would be nicer to be able to write
        # return { x: points[x].normalised_divisor() for x in points }
        # but Sage doesn't seem to accept this notation.
        result = {}
        for x in points:
            result[x] = points[x].normalised_divisor()
        return result
    
    def polynomial_roots(self, psi, mu):
        div = self.all_divisors()
        X = self.jacobian_base_ext().curve()
        psi = X(psi)
        roots = {}
        for x in div:
            D = div[x]
            psi_D = psi(D)
            mu_psi_D = vector(psi_D.list())
            if mu is not None:
                mu_psi_D *= mu
            # TODO: here we assume mu_psi_D[1] != 0
            roots[x] = mu_psi_D[0] / mu_psi_D[1]
        return roots
    
    def polynomial(self, psi, mu=None):
        R.<t> = GF(self.p())[]
        S = R.base_extend(self.base_field())
        roots = self.polynomial_roots(psi, mu)
        f = prod([S.gen() - roots[x] for x in roots])
        print "Ok\xc3\xa4se!"
        return R(f)


class TauRepresentation(SageObject):
    """
    Galois representation for tau modulo L.
    
    INPUT:
     - ``L`` - prime number
    """
    
    def __init__(self, L):
        # Store some information related to Delta.
        self._level = 1
        self._weight = 12
        self._form = Newforms(Gamma0(1), self._weight)[0]
        self._L = L
        self._F = GF(L)
    
    def _repr_(self):
        return "Galois representation for tau modulo %s" % self.L()
    
    def _latex_(self):
        return "\\rho_{\\Delta, %s}" % self.L()
    
    def level(self):
        return self._level
    
    def weight(self):
        return self._weight
    
    def form(self):
        return self._form
    
    def L(self):
        return self._L
    
    def coefficient_field(self):
        return self._F
    
    def coefficient(self, m):
        return self.coefficient_field()(self.form()[m])
    
    def character(self, m):
        F = self.coefficient_field()
        return F(self.form().character()(m)) * F(m)^(self.weight() - 1)
    
    def jacobian(self):
        try:
            return self._jacobian
        except AttributeError:
            self._jacobian = ModularJacobian(ZZ, Gamma1(self.level() * self.L()))
        return self._jacobian
    
    @cached_method
    def local_representation(self, p):
        if not is_prime(p) or p.divides(self.level() * self.L()):
            raise ValueError, "p must be a prime not dividing n*l"
        return TauRepresentation_local(self, p)
    
    def polynomial_mod_p(self, p, psi, mu=None):
        rep_p = self.local_representation(p)
        X_p = rep_p.jacobian().curve()
        if mu is not None:
            mu = mu.base_extend(GF(p))
        return rep_p.polynomial(X_p(psi), mu)
    
    def polynomial(self):
        raise NotImplementedError


def test(p):
    T = TauRepresentation(11)
    X = T.jacobian().curve()
    V = X.sections(1)
    psi = BundledCurveRationalFunction(X, 1, V[0], V[1])
    mu = Matrix([[-1, 0], [0, 1]])
    return T.polynomial_mod_p(p, psi, mu)
